# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.specs import base as spec_base
from orquesta.specs.mock import models as mock_models


VERSION = spec_base.Spec.get_version()
deserialize = mock_models.deserialize
instantiate = mock_models.instantiate
WorkflowSpec = mock_models.WorkflowSpec

__all__ = [instantiate.__name__, deserialize.__name__, WorkflowSpec.__name__]
