# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.tests.unit.composition.mistral import base


class TaskTransitionWorkflowComposerTest(base.MistralWorkflowComposerTest):
    def test_on_error(self):
        wf_name = "task-on-error"

        expected_wf_graph = {
            "directed": True,
            "graph": {},
            "nodes": [{"id": "task1"}, {"id": "task2"}, {"id": "task3"}],
            "adjacency": [
                [
                    {
                        "id": "task2",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-success"),
                    },
                    {
                        "id": "task3",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-error"),
                    },
                ],
                [],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_graph(wf_name, expected_wf_graph)

        expected_wf_ex_graph = {
            "directed": True,
            "graph": {},
            "nodes": [{"id": "task1"}, {"id": "task2"}, {"id": "task3"}],
            "adjacency": [
                [
                    {
                        "id": "task2",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-success"),
                    },
                    {
                        "id": "task3",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-error"),
                    },
                ],
                [],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_ex_graph(wf_name, expected_wf_ex_graph)

    def test_on_complete(self):
        wf_name = "task-on-complete"

        expected_wf_graph = {
            "directed": True,
            "graph": {},
            "nodes": [{"id": "task1"}, {"id": "task2"}, {"id": "task3"}, {"id": "task4"}],
            "adjacency": [
                [
                    {
                        "id": "task2",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-success"),
                    },
                    {
                        "id": "task3",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-error"),
                    },
                    {
                        "id": "task4",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-complete"),
                    },
                ],
                [],
                [],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_graph(wf_name, expected_wf_graph)

        expected_wf_ex_graph = {
            "directed": True,
            "graph": {},
            "nodes": [{"id": "task1"}, {"id": "task2"}, {"id": "task3"}, {"id": "task4"}],
            "adjacency": [
                [
                    {
                        "id": "task2",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-success"),
                    },
                    {
                        "id": "task3",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-error"),
                    },
                    {
                        "id": "task4",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-complete"),
                    },
                ],
                [],
                [],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_ex_graph(wf_name, expected_wf_ex_graph)

    def test_task_transitions_split(self):
        wf_name = "task-transitions-split"

        expected_wf_graph = {
            "directed": True,
            "graph": {},
            "nodes": [{"id": "task1"}, {"id": "task2", "splits": ["task2"]}],
            "adjacency": [
                [
                    {
                        "id": "task2",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-complete"),
                    },
                    {
                        "id": "task2",
                        "key": 1,
                        "criteria": self.compose_seq_expr("task1", condition="on-error"),
                    },
                    {
                        "id": "task2",
                        "key": 2,
                        "criteria": self.compose_seq_expr("task1", condition="on-success"),
                    },
                ],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_graph(wf_name, expected_wf_graph)

        expected_wf_ex_graph = {
            "directed": True,
            "graph": {},
            "nodes": [{"id": "task1"}, {"id": "task2", "splits": ["task2"]}],
            "adjacency": [
                [
                    {
                        "id": "task2",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-complete"),
                    },
                    {
                        "id": "task2",
                        "key": 1,
                        "criteria": self.compose_seq_expr("task1", condition="on-error"),
                    },
                    {
                        "id": "task2",
                        "key": 2,
                        "criteria": self.compose_seq_expr("task1", condition="on-success"),
                    },
                ],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_ex_graph(wf_name, expected_wf_ex_graph)
