# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta import constants


class TaskRerunRequest(object):
    def __init__(self, task_id, route, reset_items=False):
        self.task_id = task_id
        self.route = route
        self.reset_items = reset_items
        self.task_state_entry_id = constants.TASK_STATE_ROUTE_FORMAT % (task_id, str(route))
